import os
import random
from os import path

import pandas as pd
from sklearn.externals import joblib
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import LabelEncoder

from ml_platform_client.algorithm_base import AlgorithmBase, PredictorBase
from ml_platform_client.minio_accessor import MinIOAccessor
from ml_platform_client.config import global_config


class MyAlgorithm(AlgorithmBase):
    train_status_map = {}

    def train(self, model_id: str, model_path: str, data_config: dict, parameters: dict, extend: dict):
        # 拉取训练数据
        location = data_config['location']
        filename = 'temp'
        MinIOAccessor.fget_object(location, filename)

        # 数据处理
        df = pd.read_csv(filename)
        X = df[['weight', 'height']]
        label_encoder = LabelEncoder()
        y = label_encoder.fit_transform(df['gender'])

        # 模型定义
        lr = LogisticRegression()

        # 模型训练
        lr.fit(X, y)

        # 保存到磁盘
        local_model_path = model_id
        joblib.dump(lr, local_model_path)

        # 上传minio
        MinIOAccessor.save_object(local_model_path, path.join(model_path, local_model_path))

        os.remove(filename)
        os.remove(local_model_path)

    def load(self, model_id: str, model_path: str) -> PredictorBase:
        print(global_config.minio_host)
        # 从minio拉取模型文件
        full_path = path.join(model_path, model_id)
        temp_model = 'temp_model'
        MinIOAccessor.fget_object(full_path, 'temp_model')

        # 加载到内存
        lr = joblib.load(temp_model)
        os.remove(temp_model)

        # 返回predictor对象
        predictor = MyPredictor(lr)
        return predictor

    def status(self, model_id):
        return 'completed'


class MyPredictor(PredictorBase):

    def predict(self, features: dict, params: dict):
        weight = features['weight']
        height = features['height']
        result = self.model.predict([[weight, height]]).tolist()[0]
        label = None
        if result == 1:
            label = '男'
        if result == 0:
            label = '女'
        return [{'label': label, 'score': 100}]


class FakeAlgorithm(AlgorithmBase):
    train_map = {}

    def train(self, model_id: str, model_path: str, data_config: dict, parameters: dict, extend: dict):
        # 拉取训练数据
        pass

    def load(self, model_id: str, model_path: str) -> PredictorBase:
        # 从minio拉取模型文件
        # 返回predictor对象
        predictor = FakePredictor(None)
        return predictor

    def status(self, model_id):
        return 'completed'


class FakePredictor(PredictorBase):

    def predict(self, features: dict, uuid: str):
        return [{'label': random.randint(0, 10), 'score': 100}]



import os

from flask import request
from flask_restful import Resource
from .validation.validator import ValidatorBase


class BaseApi(Resource):
    api_base_url = os.environ.get('ML_PLATFORM_CLIENT_API_PREFIX', '/')

    def options(self):
        return self.get_response({}, 200)

    @staticmethod
    def check_cors(rsp):
        if 'HTTP_ORIGIN' in request.environ.keys():
            rsp.headers["Access-Control-Allow-Origin"] = request.environ['HTTP_ORIGIN']
            rsp.headers["Access-Control-Allow-Methods"] = "*"
            rsp.headers["Access-Control-Max-Age"] = "86400"
            rsp.headers["Access-Control-Allow-Headers"] = "Authorization,Origin,Content-Type,Accept"
        return rsp

    @staticmethod
    def validate(args, setup):
        for key, validator in setup.items():
            if isinstance(validator, list):
                for v in validator:
                    v.set_para_name(key)
                    v.validate(args[key])
            elif isinstance(validator, ValidatorBase):
                validator.set_para_name(key)
                validator.validate(args[key])



import uuid
from datetime import datetime, date
from functools import partial


def serialize(obj, strict=True):
    if obj is None:
        return None
    if isinstance(obj, datetime):
        return obj.strftime('%Y-%m-%d %H:%M:%S')
    if isinstance(obj, date):
        return obj.strftime('%Y-%m-%d')
    if isinstance(obj, (str, int, float, bool)):
        return obj
    if isinstance(obj, (list, tuple)):
        return list(map(partial(serialize, strict=strict), obj))
    if isinstance(obj, dict):
        new_obj = {}
        for key in obj.keys():
            new_obj[str(key)] = serialize(obj[key], strict)
        return new_obj
    if hasattr(obj, 'serialize'):
        serial = getattr(obj, 'serialize')
        if hasattr(serial, '__call__'):
            return serial()
    if strict:
        raise ValueError('object not serializable')
    return obj


def generate_uuid():
    return str(uuid.uuid4()).replace('-', '')



from collections import defaultdict

from flask import Flask
from flask_restful import Api

from .context import request_context
from .logger import log
from .dispatcher import dispatcher
from .api import TrainApi, PredictApi, LoadApi, UnloadApi, MonitorApi, StatusApi, DeleteApi

app = Flask(__name__)
api = Api(app)
api.add_resource(TrainApi, TrainApi.api_url)
api.add_resource(PredictApi, PredictApi.api_url)
api.add_resource(LoadApi, LoadApi.api_url)
api.add_resource(UnloadApi, UnloadApi.api_url)
api.add_resource(MonitorApi, MonitorApi.api_url)
api.add_resource(StatusApi, StatusApi.api_url)
api.add_resource(DeleteApi, DeleteApi.api_url)

preload_models = defaultdict(list)


def set_preload_models(alg, models):
    preload_models[alg].extend(models)


def run_app(port, debug=False):
    global app
    app.run(port=port, debug=debug)


@app.before_first_request
def load_preload_models():
    for alg in preload_models:
        for model_id in preload_models[alg]:
            dispatcher.dispatch_load(alg, model_id, None)


@app.errorhandler(Exception)
def global_handler(e):
    log.error('unhandled exception', e)
    return {'status': 'fail', 'message': str(e)}, 200


@app.teardown_request
def teardown(exception):
    request_context.uuid = None



import os
import traceback

import time
from minio import Minio


class MinIOAccessor:
    _client = None
    max_retry = 3

    def __init__(self, host, access_key, secret_key, secure):
        self.client = Minio(host, access_key=access_key, secret_key=secret_key, secure=secure)

    def fget_object(self, filepath: str, save_path: str):
        pos = filepath.index('/')
        bucket = filepath[:pos]
        filepath = filepath[pos + 1:]
        retry = 0
        while retry < MinIOAccessor.max_retry:
            try:
                self.client.fget_object(bucket, filepath, save_path)
                return
            except Exception as e:
                traceback.print_exc()
                retry += 1
                time.sleep(2)

    def save_object(self, local_path: str, remote_path: str):
        print('minio save object, local_path: {}, remote_path: {}'.format(local_path, remote_path))
        pos = remote_path.index('/')
        bucket = remote_path[:pos]
        filepath = remote_path[pos + 1:]
        with open(local_path, 'rb') as file_data:
            file_stat = os.stat(local_path)
            self.client.put_object(bucket, filepath, file_data, file_stat.st_size)

    def delete(self, bucket, prefix: str, recursive=False):
        objs = self.client.list_objects(bucket, prefix, recursive)
        for obj in objs:
            self.client.remove_object(bucket, obj.object_name)



import os

import pymysql

from .config import global_config

mysql_db_name = os.environ.get('ML_CLIENT_MLP_DB_NAME', 'ml_platform')


class MysqlAccessor:
    client = None
    # @staticmethod
    # def init_client():
    #     mysql_client = pymysql.connect(host=global_config.mysql_host, port=global_config.mysql_port,
    #                                    user=global_config.mysql_user, passwd=global_config.mysql_password,
    #                                    db=mysql_db_name)
    #     MysqlAccessor.client = mysql_client

    @staticmethod
    def get_client() -> pymysql.connections.Connection:
        return pymysql.connect(host=global_config.mysql_host, port=global_config.mysql_port,
                               user=global_config.mysql_user, passwd=global_config.mysql_password,
                               db=mysql_db_name)

    @staticmethod
    def check_load_update_time(algorithm):
        cursor = MysqlAccessor.get_client().cursor()
        cursor.execute("select load_time from load_record where algorithm = '{}'".format(algorithm))
        load_time = list(cursor.fetchall())
        return load_time[0] if load_time else None

    @staticmethod
    def get_load_models(algorithm):
        cursor = MysqlAccessor.get_client().cursor()
        cursor.execute("select model_id, path from model where algorithm = '{}' and loaded = 1".format(algorithm))
        models = list(cursor.fetchall())
        return {model_id: model_path for model_id, model_path in models}



from enum import Enum


class AlgorithmType(Enum):
    text_classification = 1
    text_sequence_labeling = 2
    text_cluster = 3
    text_generation = 4
    text_custom = 5
    relation_extraction = 6



import logging
import sys

from .context import request_context


class UuidFilter(logging.Filter):
    def filter(self, record):
        uuid = getattr(request_context, 'uuid', None)
        record.uuid = uuid
        return True


log = logging.getLogger(__name__)
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)

formatter = logging.Formatter(
    '%(asctime)s %(levelname)s %(process)d --- [%(threadName)s] %(filename)s:%(lineno)-4d: [ml-platform-client][%(uuid)s]%(message)s'
)
handler.setFormatter(formatter)

log.addHandler(handler)
log.setLevel(logging.INFO)
log.addFilter(UuidFilter())
log.propagate = False



from .util import generate_uuid


class PredictorBase:
    def __init__(self, model):
        self.model = model

    def predict(self, features: dict, params: dict):
        raise NotImplemented


class AlgorithmBase:
    def train(self, model_id: str, model_path: str, data_config: dict, parameters: dict, extend: dict):
        raise NotImplemented

    def load(self, model_id: str, model_path: str) -> PredictorBase:
        raise NotImplemented

    def unload(self, model_id):
        pass

    def status(self, model_id):
        raise NotImplemented

    def delete(self, model_id):
        raise NotImplemented

    def generate_model_id(self):
        return generate_uuid()



import random
import traceback
from collections import defaultdict
from multiprocessing import Pipe
from threading import Lock

from ml_platform_client.logger import log
from .algorithm_manager import alg_manager
from .api_util import catch_exception
from .config import global_config
from .service_response import success_response, success_response_with_data, format_service_response, error_response
from .validation.exceptions import ArgValueError, Warn
from .worker import Worker


class Dispatcher:
    def __init__(self):
        self.num_worker = global_config.num_worker
        self.num_thread = global_config.num_thread
        self.workers = []
        self.pipes = defaultdict(list)
        self.model_load_mapping = defaultdict(list)
        self.status_lock = Lock()
        for i in range(self.num_worker):
            worker_pipes = []
            for _ in range(self.num_thread):
                pipe1, pipe2 = Pipe(True)
                self.pipes[i].append((pipe1, Lock()))
                worker_pipes.append(pipe2)
            worker = Worker(worker_pipes)
            self.workers.append(worker)
            worker.start()

    def get_loaded_models(self):
        return [k for k in self.model_load_mapping if len(self.model_load_mapping[k])]

    def get_load_info(self):
        return format_service_response(success_response_with_data({'info': self.model_load_mapping}))

    def register(self, name, alg):
        alg_manager.register(name, alg)
        try:
            for worker in self.pipes:
                pipes = self.pipes[worker]
                for pipe, lock in pipes:
                    with lock:
                        pipe.send(('init_alg', alg_manager.alg_mapping))
                        pipe.recv()
        except Exception as e:
            log.error('register fail')
            log.error(e)

    def set_config(self, config):
        try:
            for worker in self.pipes:
                pipes = self.pipes[worker]
                for pipe, lock in pipes:
                    with lock:
                        pipe.send(('init_config', config))
                        pipe.recv()
        except Exception as e:
            log.error('set config fail')
            log.error(e)

    @catch_exception
    def dispatch_predict(self, model_id, features, uuid, params):
        if model_id not in self.model_load_mapping or len(self.model_load_mapping[model_id]) == 0:
            log.warn('model [{}] not loaded'.format(model_id))
            raise Warn(message='model [{}] not loaded'.format(model_id))

        candidates = self.model_load_mapping[model_id]
        index = random.choice(candidates)
        pipe, lock = random.choice(self.pipes[index])
        try:
            with lock:
                pipe.send(('predict', {'model_id': model_id, 'features': features, 'uuid': uuid, 'params': params}))
                success, predictions = pipe.recv()
                # 如果返回的是模型未加载，重置模型加载状态

            if not success and 'load' in predictions:
                with self.status_lock:
                    self.model_load_mapping[model_id].remove(index)

        except Exception as e:
            log.error(e)
            traceback.print_exc()
            log.error('fail to call worker')
            return format_service_response(error_response(message='worker fail'))

        if success:
            return format_service_response(success_response_with_data({'predictions': predictions}))
        else:
            return format_service_response(error_response(message=str(predictions)))

    @catch_exception
    def dispatch_unload(self, algorithm, model_id):
        if algorithm not in alg_manager.alg_mapping:
            raise ArgValueError(message='algorithm [{}] not support'.format(algorithm))

        if model_id not in self.model_load_mapping:
            return format_service_response(success_response('model is not loaded'))

        with self.status_lock:
            for index in self.model_load_mapping[model_id]:
                for pipe, lock in self.pipes[index]:
                    with lock:
                        pipe.send(('unload', {
                            'algorithm': algorithm,
                            'model_id': model_id
                        }))
                        success, _ = pipe.recv()

            self.model_load_mapping[model_id] = []
        return format_service_response(success_response())

    @catch_exception
    def dispatch_load(self, algorithm, model_id, model_path):
        if algorithm not in alg_manager.alg_mapping:
            raise ArgValueError(message='algorithm [{}] not support'.format(algorithm))

        if model_id in self.model_load_mapping and len(self.model_load_mapping[model_id]) > 0:
            log.info('model is already loaded {}'.format(model_id))
            return format_service_response(success_response('model is already loaded'))
            # raise ArgValueError(message='model [{}] already loaded'.format(model_id))

        index = random.randint(0, self.num_worker - 1)
        pipe, lock = random.choice(self.pipes[index])
        with lock:
            pipe.send(('load', {
                'model_id': model_id,
                'model_path': model_path,
                'algorithm': algorithm
            }))
            success, message = pipe.recv()
        if success:
            with self.status_lock:
                self.model_load_mapping[model_id].append(index)
            return format_service_response(success_response())
        else:
            return format_service_response(error_response(message=message))


dispatcher = Dispatcher()


def register(name, alg):
    global dispatcher
    dispatcher.register(name, alg())


def set_config(config):
    global dispatcher
    dispatcher.set_config(config)



from selfcheck.algorithm_type import AlgorithmType


class SelfCheckConfig:
    def __init__(self, host='127.0.0.1', algorithm_type=AlgorithmType.text_custom, algorithm_host=None,
                 algorithm_port=None, algorithm_offline_port=None, algorithm_url_prefix='', algorithm_name=None,
                 minio_host=None, minio_port=9000, minio_access_key=None, minio_secret_key=None, minio_secure=False,
                 ):
        self.host = host

        self.algorithm_type = algorithm_type
        self.algorithm_host = algorithm_host if algorithm_host else host
        self.algorithm_port = algorithm_port
        self.algorithm_offline_port = algorithm_offline_port if algorithm_offline_port else algorithm_port
        self.algorithm_url_prefix = algorithm_url_prefix
        self.algorithm_name = algorithm_name

        self.minio_host = minio_host if minio_host else host
        self.minio_port = minio_port
        self.minio_access_key = minio_access_key
        self.minio_secret_key = minio_secret_key
        self.minio_secure = minio_secure

    def validate(self):
        if not self.algorithm_name:
            raise Exception('no [algorithm_name] configured')

        if not self.algorithm_port:
            raise Exception('no [algorithm_port] configured')

        if not self.minio_access_key:
            raise Exception('no [minio_access_key] configured')

        if not self.minio_secret_key:
            raise Exception('no [minio_secret_key] configured')


class TrainConfig:
    def __init__(self, data_path=None, dict_path=None, extra_path=None, parameter=None, extend=None):
        self.data_path = data_path
        self.dict_path = dict_path
        self.extra_path = extra_path

        if extend is None:
            self.extend = {}
        else:
            self.extend = extend
        if parameter is None:
            self.parameter = {}
        else:
            self.parameter = parameter

    def validate(self):
        pass


class PredictConfig:
    def __init__(self, features=None, params=None, extend=None):
        self.features = features
        if params is None:
            self.params = {}
        else:
            self.params = params
        if extend is None:
            self.extend = {}
        else:
            self.extend = extend

    def validate(self):
        if not self.features:
            raise Exception('no [features] configured')



class ArgError(Exception):
    def __init__(self, message):
        super().__init__(message)


class ArgNoneError(ArgError):
    def __init__(self, field_name):
        self.field_name = field_name
        self.message = 'expect [{}] not null'.format(field_name)
        super().__init__(self.message)


class ArgTypeError(ArgError):
    def __init__(self, field_name, expect_type, actual_type=None):
        self.field_name = field_name
        self.expect_type = expect_type
        self.actual_type = actual_type
        if actual_type:
            self.message = 'expect [{}] to be type [{}], [{}] in actual'.format(field_name, expect_type, actual_type)
        else:
            self.message = 'expect [{}] to be type [{}]'.format(field_name, expect_type)
        super().__init__(self.message)


class ArgLengthError(ArgError):
    def __init__(self, field_name, expect_len, actual_len):
        self.field_name = field_name
        self.expect_len = expect_len
        self.actual_len = actual_len
        self.message = 'expect [{}] length less than [{}], [{}] in actual'.format(field_name, expect_len, actual_len)
        super().__init__(self.message)


class ArgRangeError(ArgError):
    def __init__(self, field_name, expect_max, expect_min, actual):
        self.field_name = field_name
        self.expect_max = expect_max
        self.expect_min = expect_min
        self.actual = actual
        if expect_max == float('inf'):
            self.message = 'expect [{}] greater than [{}], [{}] in actual'.format(field_name, expect_min, actual)
        elif expect_min == float('-inf'):
            self.message = 'expect [{}] less than [{}], [{}] in actual'.format(field_name, expect_max, actual)
        else:
            self.message = 'expect [{}] between [{}] and [{}], [{}] in actual'\
                .format(field_name, expect_min, expect_max, actual)
        super().__init__(self.message)


class ArgValueError(ArgError):
    def __init__(self, field_name=None, value=None, message=None):
        self.field_name = field_name
        if message:
            self.message = message
        else:
            self.message = 'unexpected [{}] value: {}'.format(field_name, value)
        super().__init__(self.message)


class OperationError(Exception):
    def __init__(self, message):
        super().__init__(message)


class Warn(Exception):
    def __init__(self, message):
        super().__init__(message)



import time
from threading import Thread

from .mysql_accessor import MysqlAccessor
from .algorithm_manager import alg_manager
from .dispatcher import dispatcher
from .logger import log


class Checker(Thread):

    update_time_mapping = {}

    def run(self):
        while True:
            algs = alg_manager.alg_mapping.keys()
            for algorithm in algs:
                update_time = MysqlAccessor.check_load_update_time(algorithm)
                if not update_time:
                    continue

                if algorithm not in self.update_time_mapping or update_time > self.update_time_mapping[algorithm]:
                    log.info(f'checking {algorithm}, has update')
                    models = MysqlAccessor.get_load_models(algorithm)
                    to_load_models = models.keys()
                    loaded_models = dispatcher.get_loaded_models()
                    need_load_models = to_load_models - loaded_models
                    need_unload_models = loaded_models - to_load_models

                    log.info(f'loading {algorithm} models, {need_load_models}')
                    for model_id in need_load_models:
                        dispatcher.dispatch_load(algorithm, model_id, models[model_id])

                    log.info(f'unloading {algorithm} models, {need_unload_models}')
                    for model_id in need_unload_models:
                        dispatcher.dispatch_unload(algorithm, model_id)

                    self.update_time_mapping[algorithm] = update_time
                else:
                    log.info(f'checking {algorithm}, nothing to update')

            time.sleep(5)


checker = Checker()



from os import path

from .algorithm_manager import alg_manager
from .api_util import arg_parse_validation, adapt_to_http_response, api_logger
from .base import BaseApi
from .dispatcher import dispatcher
from .setup import *


class TrainApi(BaseApi):
    api_url = path.join(BaseApi.api_base_url, 'train')

    @arg_parse_validation(arg_setup=train_arg_setup, validate_setup=train_validate_setup)
    @adapt_to_http_response
    @api_logger('train')
    def post(self, args):
        return alg_manager.train(args['algorithm'], args['model_path'], args['data'],
                                 args['parameter'], args['extend'])


class StatusApi(BaseApi):
    api_url = path.join(BaseApi.api_base_url, 'status')

    @arg_parse_validation(arg_setup=status_arg_setup, validate_setup=status_validate_setup)
    @adapt_to_http_response
    @api_logger('status')
    def get(self, args):
        return alg_manager.status(args['algorithm'], args['model_id'])


class PredictApi(BaseApi):
    api_url = path.join(BaseApi.api_base_url, 'predict')

    @arg_parse_validation(arg_setup=predict_arg_setup, validate_setup=predict_validate_setup)
    @adapt_to_http_response
    @api_logger('predict')
    def post(self, args):
        return dispatcher.dispatch_predict(args['model_id'], args['features'], args['uuid'], args['params'])


class LoadApi(BaseApi):
    api_url = path.join(BaseApi.api_base_url, 'load')

    @arg_parse_validation(arg_setup=load_arg_setup, validate_setup=load_validate_setup)
    @adapt_to_http_response
    @api_logger('load')
    def post(self, args):
        return dispatcher.dispatch_load(args['algorithm'], args['model_id'], args['model_path'])


class UnloadApi(BaseApi):
    api_url = path.join(BaseApi.api_base_url, 'unload')

    @arg_parse_validation(arg_setup=unload_arg_setup, validate_setup=unload_validate_setup)
    @adapt_to_http_response
    @api_logger('unload')
    def post(self, args):
        return dispatcher.dispatch_unload(args['algorithm'], args['model_id'])


class DeleteApi(BaseApi):
    api_url = path.join(BaseApi.api_base_url, 'delete')

    @arg_parse_validation(arg_setup=delete_arg_setup, validate_setup=delete_validate_setup)
    @adapt_to_http_response
    @api_logger('delete')
    def post(self, args):
        return alg_manager.delete(args['algorithm'], args['model_id'])


class MonitorApi(BaseApi):
    api_url = path.join(BaseApi.api_base_url, 'monitor')

    @arg_parse_validation()
    @adapt_to_http_response
    def get(self, args):
        return dispatcher.get_load_info()



from .field_type import FieldType
from .validation.validator import StringValidator, NotNullValidator, ObjectValidator

train_arg_setup = {
    'algorithm': FieldType.String,
    'model_path': FieldType.String,
    'data': FieldType.Object,
    'parameter': FieldType.Object,
    'extend': FieldType.Object,
    'uuid': FieldType.String,
}

train_validate_setup = {
    'algorithm': StringValidator(32),
    'model_path': StringValidator(1024),
    'data': [NotNullValidator()],
    'parameter': [NotNullValidator()],
    'extend': [NotNullValidator()],
}

predict_arg_setup = {
    'features': FieldType.Object,
    'model_id': FieldType.String,
    'uuid': FieldType.String,
    'params': FieldType.Object
}

predict_validate_setup = {
    'features': ObjectValidator(),
    'model_id': StringValidator(max_len=64),
}

load_arg_setup = {
    'model_path': FieldType.String,
    'model_id': FieldType.String,
    'algorithm': FieldType.String,
    'uuid': FieldType.String,
}

load_validate_setup = {
    'model_id': StringValidator(max_len=64),
    'model_path': StringValidator(max_len=1024),
    'algorithm': StringValidator(64),
}

unload_arg_setup = {
    'model_id': FieldType.String,
    'algorithm': FieldType.String,
    'uuid': FieldType.String,
}

unload_validate_setup = {
    'model_id': StringValidator(max_len=64),
    'algorithm': StringValidator(64),
}

delete_arg_setup = {
    'model_id': FieldType.String,
    'algorithm': FieldType.String,
    'uuid': FieldType.String,
}

delete_validate_setup = {
    'model_id': StringValidator(max_len=64),
    'algorithm': StringValidator(64),
}

status_arg_setup = {
    'algorithm': FieldType.String,
    'model_id': FieldType.String,
    'uuid': FieldType.String,
}

status_validate_setup = {
    'model_id': StringValidator(max_len=64),
    'algorithm': StringValidator(64),
}



import json
import time

import requests

from selfcheck.config import SelfCheckConfig
from selfcheck.constant import APP_ID_KEY
from selfcheck.executor.executor import Executor


class StatusExecutor(Executor):
    def __init__(self, config: SelfCheckConfig, model_id: str, app_id):
        self.config = config
        self.model_id = model_id
        self.url = None
        self.app_id = app_id
        self.request = None

    def prepare(self):
        self.url = 'http://{}:{}{}/{}'.format(self.config.algorithm_host, self.config.algorithm_offline_port,
                                              self.config.algorithm_url_prefix, 'status')
        self.request = {'algorithm': self.config.algorithm_name,
                        'model_id': self.model_id,
                        APP_ID_KEY: self.app_id}

    def execute(self):
        while True:
            print('status url: {}, status request: {}'.format(self.url, json.dumps(self.request, ensure_ascii=False,
                                                                                   indent=4)))
            resp = requests.get(self.url, params=self.request)
            print('status response:', resp.text)
            print()
            resp_json = resp.json()
            if resp_json['status'] == 'success':
                print('status: {}'.format(resp_json['data']['status']))
                if resp_json['data']['status'] == 'completed':
                    return
                if resp_json['data']['status'] == 'error':
                    raise Exception("status error, {}".format(json.dumps(resp_json)))

            else:
                raise Exception("status error, {}".format(json.dumps(resp_json)))
            time.sleep(5)



from .exceptions import ArgNoneError, ArgTypeError, ArgLengthError, ArgRangeError


def check_not_null(para_name, obj, is_arg=False):
    if obj is None:
        if is_arg:
            raise ArgNoneError(para_name)
        else:
            message = 'expect [{}] not null'.format(para_name)
            raise ValueError(message)


def check_type(para_name, obj, expect_type, is_arg=False):
    if not isinstance(obj, expect_type):
        if is_arg:
            raise ArgTypeError(para_name, expect_type, type(obj))
        else:
            message = 'expect [{}] to be type [{}], [{}] in actual'.format(para_name, expect_type, type(obj))
            raise TypeError(message)


def check_length(para_name, obj, expect_len):
    if len(obj) > expect_len:
        raise ArgLengthError(para_name, expect_len, len(obj))


def check_range(para_name, obj, expect_max, expect_min):
    if obj > expect_max or obj < expect_min:
        raise ArgRangeError(para_name, expect_max, expect_min, obj)



import json
import threading
import traceback
from multiprocessing import Process
from multiprocessing.connection import Connection
from typing import List

from .config import GlobalConfig, global_config
from .context import request_context
from .logger import log


class Worker(Process):
    def __init__(self, pipes: List[Connection]):
        super(Worker, self).__init__()
        self.pipes = pipes
        self.model_pool = {}
        self.alg_mapping = {}
        self.tasks = []

    def run(self):
        for pipe in self.pipes:
            self.tasks.append(threading.Thread(target=self.task, args=(pipe,)))
        for task in self.tasks:
            task.start()
        for task in self.tasks:
            task.join()

    def task(self, pipe):
        while True:
            try:
                command, data = pipe.recv()
                if command == 'exit':
                    return
                try:
                    result = self._process(command, data)
                except Exception as e:
                    result = False, None
                    log.error('worker process error: {}, {}'.format(command, e))
                    traceback.print_exc()
                pipe.send(result)
            except Exception as e:
                log.error('worker communication error, {}'.format(e))
                traceback.print_exc()

    def _process(self, command, data):
        if command == 'load':
            return self._process_load(data)
        if command == 'unload':
            return self._process_unload(data)
        if command == 'predict':
            return self._process_predict(data)
        if command == 'init_alg':
            return self._process_init_alg(data)
        if command == 'init_config':
            return self._process_init_config(data)
        return False, None

    def _process_init_alg(self, data):
        self.alg_mapping.update(data)

    def _process_init_config(self, data: GlobalConfig):
        global_config.minio_host = data.minio_host
        global_config.minio_access_key = data.minio_access_key
        global_config.minio_secret_key = data.minio_secret_key
        global_config.minio_secure = data.minio_secure
        return True, None

    def _process_load(self, data):
        model_id = data['model_id']
        model_path = data['model_path']
        algorithm = data['algorithm']
        if algorithm not in self.alg_mapping:
            return False, 'algorithm not support'

        try:
            model = self.alg_mapping[algorithm].load(model_id, model_path)
            self.model_pool[model_id] = model
            return True, None
        except Exception as e:
            log.error('load fail', e)
            log.error(traceback.format_exc())
            return False, 'load fail {}'.format(e)

    def _process_unload(self, data):
        try:
            model_id = data['model_id']
            algorithm = data['algorithm']
            self.alg_mapping[algorithm].unload(model_id)
            if model_id in self.model_pool:
                del self.model_pool[model_id]
            return True, None
        except Exception as e:
            log.error('unload fail', e)
            log.error(traceback.format_exc())
            return False, 'unload fail {}'.format(e)

    def _process_predict(self, data):
        model_id = data['model_id']
        features = data['features']
        params = data['params']
        uuid = data['uuid']
        request_context.uuid = uuid
        if model_id not in self.model_pool:
            message = 'model [{}] not loaded'.format(model_id)
            log.warn(message)
            return False, message

        try:
            prediction = self.model_pool[model_id].predict(features, params)
            log.info('[{}]model: {}, features: {}, prediction: {}, params: {}'
                     .format(uuid, model_id,
                             json.dumps(features, ensure_ascii=False),
                             json.dumps(prediction, ensure_ascii=False),
                             json.dumps(prediction, ensure_ascii=False)))

            return True, prediction
        except Exception as e:
            log.error('predict fail', e)
            log.error(traceback.format_exc())
            return False, 'predict fail {}'.format(e)
        finally:
            request_context.uuid = None



import os
import traceback

import time
from minio import Minio

from .config import global_config
from .logger import log


class MinIOAccessor:
    _client = None
    max_retry = 3

    @staticmethod
    def init_client(host, access_key, secret_key, secure):
        minio_client_instance = Minio(host, access_key=access_key, secret_key=secret_key, secure=secure)
        log.info('minio connection info: host={}, access_key={}, secret_key={}, secure={}'
                 .format(host, access_key, secret_key, secure))
        MinIOAccessor._client = minio_client_instance

    @staticmethod
    def get_client() -> Minio:
        if MinIOAccessor._client is None:
            # minio_host = '172.16.102.156:9000'
            # minio_access_key = '24WTCKX23M0MRI6BA72Y'
            # minio_secret_key = '6mxoQMf1Lha2q1H3fp3fjNHEZRJJ7LvbC5k3+CQX'
            # minio_secure = False
            MinIOAccessor.init_client(global_config.minio_host, global_config.minio_access_key,
                                      global_config.minio_secret_key, global_config.minio_secure)
            # raise Exception('client is not initialized')
        return MinIOAccessor._client

    @staticmethod
    def fget_object(filepath: str, save_path: str):
        pos = filepath.index('/')
        bucket = filepath[:pos]
        filepath = filepath[pos + 1:]
        retry = 0
        while retry < MinIOAccessor.max_retry:
            try:
                MinIOAccessor.get_client().fget_object(bucket, filepath, save_path)
                log.info('read file from minio success: {}.{}'.format(bucket, filepath, retry))
                return
            except Exception as e:
                log.error('read file from minio error: {}.{}, retry:{}'.format(bucket, filepath, retry))
                log.error(e)
                traceback.print_exc()
                retry += 1
                time.sleep(2)
        log.error('read file from minio fail: {}.{}'.format(bucket, filepath))

    @staticmethod
    def save_object(local_path: str, remote_path: str):
        pos = remote_path.index('/')
        bucket = remote_path[:pos]
        filepath = remote_path[pos + 1:]
        with open(local_path, 'rb') as file_data:
            file_stat = os.stat(local_path)
            MinIOAccessor.get_client().put_object(bucket, filepath, file_data, file_stat.st_size)

    @staticmethod
    def delete(bucket, prefix: str, recursive=False):
        objs = MinIOAccessor.get_client().list_objects(bucket, prefix, recursive)
        for obj in objs:
            MinIOAccessor.get_client().remove_object(bucket, obj.object_name)



from multiprocessing import Process, Pipe
from multiprocessing.connection import Connection


def ping(conn):
    for i in range(3):
        conn.send('pong')
        msg = conn.recv()
        print(msg)
    conn.send('exit')


def pong(conn):
    while True:
        msg = conn.recv()
        if msg == 'exit':
            print('exit')
            return
        print(msg)
        conn.send('ping')


class Worker1(Process):
    def __init__(self, pipe: Connection):
        super(Worker1, self).__init__()
        self.pipe = pipe
        self.model_pool = {}
        self.alg_mapping = {}

    def run(self):
        try:
            while True:
                try:
                    command, data = self.pipe.recv()
                    if command == 'exit':
                        return
                    try:
                        result = command + ' success'
                    except Exception as e:
                        result = False, None

                    self.pipe.send(result)
                except Exception as e:
                    print(e)
        except Exception as e:
            print(e)


if __name__ == '__main__':
    parent_conn, child_conn = Pipe()
    worker1 = Worker1(child_conn)
    worker1.start()
    parent_conn.send(('cmd1', []))
    parent_conn.send(('cmd2', []))
    parent_conn.send(('exit', []))
    print(parent_conn.recv())
    print(parent_conn.recv())
    print('over')



from .exceptions import ArgValueError
from .para import check_type, check_length, check_range, check_not_null


class ValidatorBase:
    def __init__(self):
        self.para_name = 'field'

    def validate(self, value):
        raise NotImplementedError()

    def set_para_name(self, name):
        self.para_name = name


class NotNullValidator(ValidatorBase):
    def validate(self, value):
        return check_not_null(self.para_name, value, is_arg=True)


class TypeValidator(ValidatorBase):
    def __init__(self, value_type):
        super().__init__()
        self.value_type = value_type

    def validate(self, value):
        return not value or check_type(self.para_name, value, self.value_type, is_arg=True)


class LengthValidator(ValidatorBase):
    def __init__(self, max_len):
        super().__init__()
        self.max_len = max_len

    def validate(self, value):
        return not value or check_length(self.para_name, value, self.max_len)


class RangeValidator(ValidatorBase):
    def __init__(self, max_value=float('inf'), min_value=float('-inf')):
        super().__init__()
        self.max_value = max_value
        self.min_value = min_value

    def validate(self, value):
        return not value or check_range(self.para_name, value, self.max_value, self.min_value)


class StringValidator(ValidatorBase):
    def __init__(self, max_len=None, nullable=False):
        super().__init__()
        self.validators = [TypeValidator(str)]
        if not nullable:
            self.validators.append(NotNullValidator())
        if max_len:
            self.validators.append(LengthValidator(max_len))

    def validate(self, value):
        for validator in self.validators:
            validator.set_para_name(self.para_name)
            validator.validate(value)


class ObjectValidator(ValidatorBase):
    def __init__(self, primitive_types=(bool, int, float, str), allow_list=True, allow_dict=False):
        super().__init__()
        self.type_validator = TypeValidator(dict)
        self.type_validator.set_para_name(self.para_name)
        self.primitive_types = primitive_types
        self.allow_list = allow_list
        self.allow_dict = allow_dict

    def validate(self, value):
        self.type_validator.validate(value)
        if value:
            for key in value.keys():
                self.validate_value(key, value[key])

    def validate_value(self, key, value):
        if isinstance(value, self.primitive_types):
            return
        if isinstance(value, list):
            if self.allow_list:
                for v in value:
                    self.validate_value(key, v)
                return
            else:
                raise ArgValueError(message='expect [{}] not contain list for key [{}]'.format(self.para_name, key))
        if isinstance(value, dict):
            if self.allow_dict:
                for key in value.keys():
                    self.validate_value(key, value[key])
                return
            else:
                raise ArgValueError(message='expect [{}] not contain dict for key [{}]'.format(self.para_name, key))
        raise ArgValueError(message='unexpect type for key [{}] in [{}]'.format(key, self.para_name))


class StringObjectValidator(ObjectValidator):
    def __init__(self):
        super().__init__(primitive_types=(str,), allow_list=False, allow_dict=False)



from ml_platform_client.config import global_config
from ml_platform_client.dispatcher import register, set_config
from ml_platform_client.server import app
from ml_platform_client.server import set_preload_models
from example.basic.algorithms import MyAlgorithm, FakeAlgorithm

minio_host = '172.16.99.29:9000'
minio_access_key = '24WTCKX23M0MRI6BA72Y'
minio_secret_key = '6mxoQMf1Lha2q1H3fp3fjNHEZRJJ7LvbC5k3+CQX'
minio_secure = False
mysql_host = '172.16.99.29'
mysql_port = 3306
mysql_user = 'root'
mysql_password = 'password'

if __name__ == '__main__':
    global_config.minio_host = minio_host
    global_config.minio_access_key = minio_access_key
    global_config.minio_secret_key = minio_secret_key
    global_config.minio_secure = minio_secure

    global_config.mysql_host = mysql_host
    global_config.mysql_port = mysql_port
    global_config.mysql_user = mysql_user
    global_config.mysql_password = mysql_password

    register('my_alg', MyAlgorithm)
    register('fake_alg', FakeAlgorithm)
    set_preload_models('my_alg', ['pretrain_model_id1', 'pretrain_model_id2'])

    # checker.start()

    set_config(global_config)
    app.run(host='0.0.0.0', port=5000)



import json

from selfcheck.algorithm_type import AlgorithmType
from selfcheck.config import SelfCheckConfig, TrainConfig, PredictConfig
from selfcheck.constant import APP_ID_KEY
from selfcheck.executor.train_executor import TrainExecutor
from selfcheck.executor.predict_executor import PredictExecutor
from selfcheck.executor.unload_executor import UnloadExecutor
from selfcheck.executor.load_executor import LoadExecutor
from selfcheck.executor.status_executor import StatusExecutor


class SelfCheckRunner:
    def __init__(self, config: SelfCheckConfig):
        self.config = config
        self.config.validate()

    def run_train(self, train_config: TrainConfig):
        model_id = self._execute_train(train_config)
        print('train started, model_id: {}'.format(model_id))
        print()

        self._execute_status(model_id, train_config.extend.get(APP_ID_KEY, None))
        print('train done')
        print()
        return model_id

    def run_predict(self, model_id: str, predict_config: PredictConfig):
        self._execute_load(model_id, predict_config)
        print('load done')
        print()

        predictions = self._execute_predict(model_id, predict_config)
        print('prediction done:')
        print(json.dumps(predictions, ensure_ascii=False, indent=4))
        self.validate_prediction(predictions)
        print()

        self._execute_unload(model_id, predict_config)
        print('unload done')
        print()

    def run_all(self, train_config: TrainConfig, predict_config: PredictConfig):
        model_id = self.run_train(train_config)
        self.run_predict(model_id, predict_config)

    def _execute_train(self, train_config: TrainConfig):
        return self._execute(TrainExecutor(self.config, train_config))

    def _execute_status(self, model_id, app_id):
        return self._execute(StatusExecutor(self.config, model_id, app_id))

    def _execute_load(self, model_id, predict_config):
        return self._execute(LoadExecutor(self.config, model_id, predict_config.extend))

    def _execute_predict(self, model_id, predict_config: PredictConfig):
        return self._execute(PredictExecutor(self.config, model_id, predict_config))

    def _execute_unload(self, model_id, predict_config):
        return self._execute(UnloadExecutor(self.config, model_id, predict_config.extend))

    @staticmethod
    def _execute(executor):
        executor.prepare()
        return executor.execute()

    def validate_prediction(self, predictions):
        if self.config.algorithm_type == AlgorithmType.text_classification or self.config.algorithm_type == AlgorithmType.text_generation:
            self.validate_text_classification(predictions)

        if self.config.algorithm_type == AlgorithmType.text_sequence_labeling:
            self.validate_text_sequence_labelilng(predictions)

        if self.config.algorithm_type == AlgorithmType.text_cluster:
            self.validate_text_cluster(predictions)

    def validate_text_cluster(self, predictions):
        assert isinstance(predictions, list)
        for prediction in predictions:
            assert 'score' in prediction
            self.validate_score(prediction['score'])

            assert 'cluster_name' in prediction
            assert isinstance(prediction['cluster_name'], str)

            assert 'cluster' in prediction
            assert isinstance(prediction['cluster'], list)
            for cluster in prediction['cluster']:
                assert isinstance(cluster, str)

            if 'other_info' in prediction:
                self.validate_other_info(prediction['other_info'])

    def validate_text_sequence_labelilng(self, predictions):
        assert isinstance(predictions, list)
        for prediction in predictions:
            assert 'score' in prediction
            self.validate_score(prediction['score'])

            assert 'entity_list' in prediction
            assert isinstance(prediction['entity_list'], list)

            entity = prediction['entity_list'][0]
            assert 'name' in entity
            assert isinstance(entity['name'], str)

            assert 'start_pos' in entity
            assert isinstance(entity['start_pos'], int)

            assert 'end_pos' in entity
            assert isinstance(entity['end_pos'], int)

            assert 'label' in entity
            assert isinstance(entity['label'], str)

            assert 'score' in entity
            assert self.validate_score(entity['score'])

    def validate_text_classification(self, predictions):
        assert isinstance(predictions, list)
        for prediction in predictions:
            assert 'label' in prediction
            assert isinstance(prediction['label'], str)

            assert 'score' in prediction
            self.validate_score(prediction['score'])

            if 'other_info' in prediction:
                self.validate_other_info(prediction['other_info'])

    def validate_relation_extraction(self, predictions):
        assert isinstance(predictions, list)
        for prediction in predictions:
            assert 'score' in prediction
            assert self.validate_score(prediction['score'])

            assert 'subject' in prediction
            self.validate_entity(prediction['subject'])

            assert 'object' in prediction
            self.validate_entity(prediction['object'])

            assert 'predicate' in prediction
            self.validate_predicate(prediction['predicate'])

    def validate_predicate(self, predicate):
        assert 'text' in predicate
        assert isinstance(predicate['text'], str)

        assert 'mention' in predicate
        assert isinstance(predicate['mention'], list)
        for mention in predicate['mention']:
            assert 'text' in mention
            assert isinstance(mention['text'], str)

            assert 'start' in mention
            assert isinstance(mention['start'], int)

            assert 'end' in mention
            assert isinstance(mention['end'], int)

            assert 'pos' in mention
            assert isinstance(mention['pos'], list)

        if 'property' in predicate:
            assert isinstance(predicate['property'], list)
            for prop in predicate['property']:
                assert 'label' in prop
                assert isinstance(prop['label'], str)

                assert 'value' in prop
                assert isinstance(prop['value'], list)

                for value in prop['value']:
                    assert 'text' in value
                    assert isinstance(value['text'], str)

                    assert 'start' in value
                    assert isinstance(value['start'], int)

                    assert 'end' in value
                    assert isinstance(value['end'], int)

                    assert 'pos' in value
                    assert isinstance(value['pos'], str)

    def validate_entity(self, entity):
        assert 'name' in entity
        assert isinstance(entity['name'], str)

        assert 'start' in entity
        assert isinstance(entity['start'], int)

        assert 'end' in entity
        assert isinstance(entity['end'], int)

        assert 'pos' in entity
        assert isinstance(entity['pos'], str)

        assert 'label' in entity
        assert isinstance(entity['label'], str)

    def validate_other_info(self, other_infos):
        assert isinstance(other_infos, list)
        for other_info in other_infos:
            assert 'key' in other_info
            assert isinstance(other_info['key'], str)

            assert 'value' in other_info

    def validate_score(self, score):
        return isinstance(score, float) or isinstance(score, int)



import json

import requests

from selfcheck.config import SelfCheckConfig, TrainConfig
from selfcheck.executor.executor import Executor
from selfcheck.minio_accessor import MinIOAccessor
from selfcheck.util import get_uuid


class TrainExecutor(Executor):

    def __init__(self, config: SelfCheckConfig, train_config: TrainConfig):
        self.config = config
        self.train_config = train_config
        self.train_request = None
        self.url = None
        self.model_id = None
        minio_addr = '{}:{}'.format(config.minio_host, config.minio_port)
        self.minio_client = MinIOAccessor(minio_addr, config.minio_access_key, config.minio_secret_key,
                                          config.minio_secure)

    def prepare(self):
        self.model_id = get_uuid()
        self.train_request = {
            'model_id': self.model_id,
            'algorithm': self.config.algorithm_name,
            'parameter': self.train_config.parameter,
            'model_path': '',
            'extend': self.train_config.extend,
            'data': {
                'location': None
            }
        }
        self.url = 'http://{}:{}{}/{}'.format(self.config.algorithm_host, self.config.algorithm_offline_port,
                                              self.config.algorithm_url_prefix, 'train')

        if self.train_config.data_path:
            data_path = 'test/' + get_uuid()
            minio_path = 'ml-platform-data/' + data_path
            self.minio_client.save_object(self.train_config.data_path, minio_path)
            self.train_request['data']['location'] = data_path

        if self.train_config.dict_path:
            dict_path = 'test/' + get_uuid()
            minio_path = 'ml-platform-data/' + dict_path
            self.minio_client.save_object(self.train_config.dict_path, minio_path)
            self.train_request['extend']['dict_location'] = dict_path

        if self.train_config.extra_path:
            extra_path = 'test/' + get_uuid()
            minio_path = 'ml-platform-data/' + extra_path
            self.minio_client.save_object(self.train_config.extra_path, minio_path)
            self.train_request['extend']['extra_location'] = extra_path

    def execute(self):
        print('train url: {}, train request: {}'.format(self.url, json.dumps(self.train_request, ensure_ascii=False,
                                                                             indent=4)))
        resp = requests.post(self.url, json=self.train_request)
        print('train response:', resp.text)
        print()
        resp_json = resp.json()
        if resp_json['status'] == 'success':
            if resp_json['data']['model_id']:
                return resp_json['data']['model_id']
            else:
                return self.model_id
        else:
            raise Exception('train error, {}'.format(json.dumps(resp_json)))



import json

import requests

from selfcheck.config import SelfCheckConfig
from selfcheck.executor.executor import Executor


class LoadExecutor(Executor):
    def __init__(self, config: SelfCheckConfig, model_id: str, extend=None):
        self.config = config
        self.model_id = model_id
        self.request = None
        self.url = None
        if extend is None:
            self.extend = {}
        else:
            self.extend = extend

    def prepare(self):
        self.request = {
            'algorithm': self.config.algorithm_name,
            'model_id': self.model_id,
            'extend': self.extend,
            'model_path': ''
        }
        self.url = 'http://{}:{}{}/{}'.format(self.config.algorithm_host, self.config.algorithm_port,
                                              self.config.algorithm_url_prefix, 'load')

    def execute(self):
        print('load url: {}, load request: {}'.format(self.url, json.dumps(self.request, ensure_ascii=False,
                                                                           indent=4)))
        resp = requests.post(self.url, json=self.request)
        print('load response:', resp.text)
        resp_json = resp.json()
        if resp_json['status'] == 'success':
            return
        else:
            raise Exception('status error, {}'.format(json.dumps(resp_json)))



import os


class GlobalConfig:
    def __init__(self):
        self.server_url = None
        self.minio_host = None
        self.minio_access_key = None
        self.minio_secret_key = None
        self.minio_secure = None
        self.mysql_host = None
        self.mysql_port = None
        self.mysql_user = None
        self.mysql_password = None
        self.num_worker = int(os.environ.get('MLP_CLIENT_WORKER', 5))
        self.num_thread = int(os.environ.get('MLP_CLIENT_THREAD', 10))


global_config = GlobalConfig()



import logging
from multiprocessing import Manager

from flask import request

from .service_response import success_response_with_data, format_service_response, success_response
from .api_util import catch_exception
from .validation.exceptions import ArgValueError


class AlgorithmManager:
    def __init__(self):
        manager = Manager()
        self.alg_mapping = manager.dict()

    def register(self, name, algorithm):
        self.alg_mapping[name] = algorithm

    @catch_exception
    def train(self, algorithm, model_path, data_config, parameters, extend):
        if algorithm not in self.alg_mapping:
            raise ArgValueError(message='algorithm [{}] not support'.format(algorithm))
        extend['remote_ip'] = request.remote_addr

        model_id = self.alg_mapping[algorithm].generate_model_id()
        result = self._train_async(algorithm, model_id, model_path, data_config, parameters, extend)
        if result:
            model_id = result
        return format_service_response(success_response_with_data({'model_id': model_id}))

    def _train_async(self, algorithm, model_id, model_path, data_config, parameters, extend):
        try:
            result = self.alg_mapping[algorithm].train(model_id, model_path, data_config, parameters, extend)
            self._train_call_back(model_id, result)
            logging.info('train model [{}] success'.format(model_id))
            return result
        except Exception as e:
            logging.error(e)
            logging.error('train model [{}] failed'.format(model_id))
            raise e

    def _train_call_back(self, model_id, success):
        pass

    @catch_exception
    def status(self, algorithm, model_uid):
        status = self.alg_mapping[algorithm].status(model_uid)
        return format_service_response(success_response_with_data({'status': status}))

    @catch_exception
    def delete(self, algorithm, model_uid):
        self.alg_mapping[algorithm].delete(model_uid)
        return format_service_response(success_response())


alg_manager = AlgorithmManager()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ml-platform-client",
    version="0.3.18.2",
    author="JeremyXin",
    author_email="chengjiexin@emotibot.com",
    description="emotibot ml platform client for python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url="https://github.com/pypa/sampleproject",
    packages=setuptools.find_packages(),
    install_requires=[
        "flask==1.1.1",
        "flask-restful==0.3.8",
        "minio==5.0.5",
        "pymysql==0.7.11"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import json
import traceback
from datetime import datetime, date
from functools import wraps

import time
from flask import Response
from flask_restful.reqparse import RequestParser
from werkzeug.datastructures import FileStorage

from .context import request_context
from .field_type import FieldType
from .logger import log
from .validation.exceptions import ArgError, ArgTypeError
from .validation.validator import ValidatorBase, TypeValidator


class JsonSerializable(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(obj, date):
            return obj.strftime('%Y-%m-%d')
        else:
            return json.JSONEncoder.default(self, obj)


type_mapping = {
    FieldType.Int: int,
    FieldType.Float: float,
    FieldType.String: str,
    FieldType.Bool: bool,
    FieldType.Array: list,
    FieldType.Object: dict,
}


def parse_args(rules):
    parser = RequestParser()
    try:
        for key, value in rules.items():
            if value == FieldType.File:
                parser.add_argument(key, type=FileStorage, location='files')
            elif value == FieldType.Array:
                parser.add_argument(key, action='append')
            elif value in type_mapping:
                parser.add_argument(key, type=type_mapping[value])
        return parser.parse_args()
    except Exception as e:
        if hasattr(e, 'data') and 'message' in e.data:
            key = list(e.data['message'].keys())[0]
            raise ArgTypeError(key, type_mapping[rules[key]])
        else:
            raise e


def validate(args, setup):
    for key, validator in setup.items():
        if isinstance(validator, list):
            for v in validator:
                v.set_para_name(key)
                v.validate(args[key])
        elif isinstance(validator, ValidatorBase):
            validator.set_para_name(key)
            validator.validate(args[key])


def arg_parse_validation(arg_setup=None, validate_setup=None):
    if arg_setup is None:
        arg_setup = {}
    if validate_setup is None:
        validate_setup = {}
    arg_setup['trace'] = FieldType.Bool
    arg_setup['debug'] = FieldType.Bool
    validate_setup['trace'] = TypeValidator(bool)
    validate_setup['debug'] = TypeValidator(bool)

    def outer(func):
        @wraps(func)
        def inner(*args, **kwargs):
            try:
                parsed_args = parse_args(arg_setup)
                if 'uuid' in parsed_args:
                    request_context.uuid = parsed_args['uuid']

                validate(parsed_args, validate_setup)
                kwargs['args'] = parsed_args
                return func(*args, **kwargs)
            except ArgError as e:
                res, status = handle_exception(e, func)
                log.warn('arg error, response: {}'.format(res))
                return Response(json.dumps(res, cls=JsonSerializable), mimetype="application/json", status=200)

        return inner

    return outer


def adapt_to_http_response(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            res = func(*args, **kwargs)
        except Exception as e:
            res, status = handle_exception(e, func)
        return Response(json.dumps(res, cls=JsonSerializable), mimetype="application/json", status=200)

    return wrapper


def handle_exception(e, func):
    message = str(e)
    if isinstance(e, ArgError):
        api_status = 'arg_error'
        res = {'status': api_status, 'message': message}
    else:
        api_status = 'fail'
        res = {'status': api_status, 'message': message, 'trace': traceback.format_exc()}
    log.error("exception in {}.{}: {}".format(str(func), func.__name__, message))
    traceback.print_exc()
    return res, 200


def catch_exception(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            message = str(e)
            if isinstance(e, ArgError):
                api_status = 'arg_error'
                res = {'status': api_status, 'message': message}
            else:
                api_status = 'fail'
                res = {'status': api_status, 'message': message, 'trace': traceback.format_exc()}
            log.error("exception in {}.{}: {}".format(str(func), func.__name__, message))
            traceback.print_exc()
            return res

    return wrapper


def api_logger(method_name):
    def outer(func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            api_args = kwargs['args']
            start = time.time()
            res = func(*args, **kwargs)
            cost = time.time() - start
            log.info('method: {}, args: {}, response: {}, cost: {}'.format(method_name,
                                                                           json.dumps(api_args, ensure_ascii=False),
                                                                           json.dumps(res, ensure_ascii=False),
                                                                           cost))
            return res

        return wrapper

    return outer



from selfcheck.algorithm_type import AlgorithmType
from selfcheck.config import SelfCheckConfig, TrainConfig, PredictConfig
from selfcheck.self_check_runner import SelfCheckRunner

if __name__ == '__main__':
    config = SelfCheckConfig(host='172.16.99.29', algorithm_name='ft_service', algorithm_port=8895,
                             algorithm_offline_port=8896,
                             algorithm_type=AlgorithmType.text_classification,
                             minio_access_key='24WTCKX23M0MRI6BA72Y',
                             minio_secret_key='6mxoQMf1Lha2q1H3fp3fjNHEZRJJ7LvbC5k3+CQX')

    train_config = TrainConfig(data_path='./static/classify_train.csv')
    predict_config = PredictConfig(features={'sentence': '什么情况下，需要进行居家隔离'})

    task_runner = SelfCheckRunner(config)
    task_runner.run_all(train_config, predict_config)



from collections import namedtuple

from .util import serialize

ServiceResponse = namedtuple('ServiceResult', ['success', 'status', 'data', 'message',
                                               'trace', 'need_data'])


def success_response(message=''):
    return ServiceResponse(success=True, status='success', data=None, message=message,
                           trace='', need_data=False)


def success_response_with_data(data=None, message=''):
    return ServiceResponse(success=True, status='success', data=data, message=message,
                           trace='', need_data=True)


def error_response(status='fail', message='', trace=''):
    return ServiceResponse(success=False, status=status, data=None, message=message,
                           trace=trace, need_data=False)


def error_response_with_data(status='fail', data=None, message='', trace='', data_key='data'):
    assert data_key != 'status'
    assert data_key != 'message'
    assert data_key != 'trace'
    return ServiceResponse(success=False, status=status, data=data, message=message,
                           trace=trace, need_data=True)


def format_service_response(response):
    if response.status == 'success':
        return _respond_success(response)
    if response.status == 'fail':
        return _respond_fail(response)
    return _respond_other(response)


def _respond_fail(response):
    return {
        'status': response.status,
        'message': response.message,
        'trace': response.trace
    }


def _respond_other(response):
    resp = {
        'status': response.status
    }
    _append_data(resp, response)
    if response.message:
        resp['message'] = response.message
    return resp


def _respond_success(response):
    resp = {
        'status': response.status
    }
    _append_data(resp, response)
    return resp


def _append_data(resp, response):
    if response.need_data:
        resp['data'] = serialize(response.data)
